 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
	History:
		- written by W. Wehner,  June 1995
*/

#include <stdio.h>
#include <stdlib.h>

#include <GL/gl.h>

#define XF 7
#define YF 7
#define	ABS(a) ((a)<0 ? -(a) : (a))
#define	SUP(a,b) ((a)>(b) ? (a) : (b))
#define	INF(a,b) ((a)<(b) ? (a) : (b))

int MakeFilterKernel( GLfloat, GLfloat, int );
static void ResetContext( void );
static void MakeEdgeKernel( GLfloat, GLfloat );
static void MakeBlurKernel( GLfloat, GLfloat );
static void MakeSepBlurKernel( GLfloat, GLfloat );
static void MakeSharpenKernel( GLfloat, GLfloat );
static void MakeScaleBias( GLfloat, GLfloat );

/*
	returns shrink size of pixel transfers
*/
int
MakeFilterKernel( GLfloat xfactor, GLfloat yfactor, int which)
{
	ResetContext();
	switch (which%4)
	{
		case 0: MakeScaleBias(xfactor, yfactor); return 0;
		case 1: MakeEdgeKernel(xfactor, yfactor); return 2;
		case 2: MakeSharpenKernel(xfactor, yfactor); return 2;
		case 3: MakeBlurKernel(xfactor, yfactor); return (XF-1);
		case 4: MakeSepBlurKernel(xfactor, yfactor); return (XF-1);
	}
}

static void
ResetContext( void )
{
    glPixelTransferf(GL_RED_SCALE, 1.0);
    glPixelTransferf(GL_RED_BIAS, 0.0);
			
    glPixelTransferf(GL_GREEN_SCALE, 1.0);
    glPixelTransferf(GL_GREEN_BIAS, 0.0);
			
    glPixelTransferf(GL_BLUE_SCALE, 1.0);
    glPixelTransferf(GL_BLUE_BIAS, 0.0);
			
    glPixelTransferf(GL_ALPHA_SCALE, 1.0);
    glPixelTransferf(GL_ALPHA_BIAS, 0.0);

    glDisable(GL_CONVOLUTION_2D_EXT);
    glDisable(GL_SEPARABLE_2D_EXT);
}

static void
MakeBlurKernel(GLfloat xfactor, GLfloat yfactor)
{
#if 0
    float kernel[XF][YF][4];
    int i, j, k;
	int xf, yf;
    float weight;

    weight = 1./((float)XF*(float)YF);
		    
    for (i=0;i<XF;i++)
    for (j=0;j<YF;j++)
	    for(k=0;k<4;k++)
			kernel[i][j][k] = weight;
			    
    glEnable(GL_CONVOLUTION_2D_EXT);
    glConvolutionFilter2DEXT(GL_CONVOLUTION_2D_EXT, GL_RGBA, XF, YF,
			     GL_RGBA, GL_FLOAT, kernel );
#else
    float kernel[XF][YF];
    int i, j;
	int xf, yf;
    float weight;

    weight = 1./((float)XF*(float)YF);
		    
    for (i=0;i<XF;i++)
    for (j=0;j<YF;j++)
			kernel[i][j] = weight;
			    
    glEnable(GL_CONVOLUTION_2D_EXT);
    glConvolutionFilter2DEXT(GL_CONVOLUTION_2D_EXT, GL_LUMINANCE, XF, YF,
			     GL_LUMINANCE, GL_FLOAT, kernel );
#endif
}

static void
MakeSepBlurKernel(GLfloat xfactor, GLfloat yfactor)
{
    float row[XF][4], column[YF][4];
    int i, j, k;
	int xf, yf;
    float xweight, yweight;

    xweight = 1./(float)XF;
    yweight = 1./(float)YF;
		    
    for (i=0;i<XF;i++)
	    for(k=0;k<4;k++)
			row[i][k] = xweight;
	for(i=0;i<YF;i++)
	    for(k=0;k<4;k++)
			column[i][k] = yweight;
			    
    glEnable(GL_SEPARABLE_2D_EXT);
    glSeparableFilter2DEXT(GL_SEPARABLE_2D_EXT, GL_RGBA, XF, YF,
			     GL_RGBA, GL_FLOAT, row, column);
}

static void
MakeEdgeKernel(GLfloat xfactor, GLfloat yfactor)
{
    float kernel[3][3][4];
    int i, j, k;

    for (i=0;i<3;i++)
	for(j=0;j<3;j++)
	    for(k=0;k<4;k++)
		kernel[i][j][k] = (i-1)*yfactor + (j-1)*xfactor;
			    
    glEnable(GL_CONVOLUTION_2D_EXT);
    glConvolutionFilter2DEXT(GL_CONVOLUTION_2D_EXT,GL_RGBA, 3, 3, GL_RGBA,
			     GL_FLOAT, kernel);
}

static void
MakeSharpenKernel(GLfloat xfactor, GLfloat yfactor)
{
    float kernel[3][3][4];
    int i, j, k;

    for (i=0;i<3;i++)
	for(j=0;j<3;j++)
	    for(k=0;k<4;k++)
		kernel[i][j][k] = -(xfactor + yfactor);
			
    for(k=0;k<4;k++)
	kernel[1][1][k] = (xfactor+yfactor)*8+1;
			    
    glEnable(GL_CONVOLUTION_2D_EXT);
    glConvolutionFilter2DEXT(GL_CONVOLUTION_2D_EXT,GL_RGBA, 3, 3, GL_RGBA,
			     GL_FLOAT,  kernel);
}

static void
MakeScaleBias(GLfloat xfactor, GLfloat yfactor)
{
    glPixelTransferf(GL_RED_SCALE, xfactor);
    glPixelTransferf(GL_RED_BIAS, yfactor);
			
    glPixelTransferf(GL_GREEN_SCALE, xfactor);
    glPixelTransferf(GL_GREEN_BIAS, yfactor);
			
    glPixelTransferf(GL_BLUE_SCALE, xfactor);
    glPixelTransferf(GL_BLUE_BIAS, yfactor);
			
    glPixelTransferf(GL_ALPHA_SCALE, xfactor);
    glPixelTransferf(GL_ALPHA_BIAS, yfactor);
}
