 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <GL/glut.h>

#include <gl/image.h>

#define	INF(a,b)	((a)<(b) ? (a) : (b))
#define	SUP(a,b)	((a)>(b) ? (a) : (b))
#define	DEG2RAD(a)	((a)*M_PI/180.)

#define	ARROW_STRIDE	5.
#define	RIGHTWARD	1
#define	LEFTWARD	2
#define	UPWARD		3
#define	DOWNWARD	4

#ifdef SUBTEX
#define	TILE_S_SIZE	64
#define	TILE_T_SIZE	64
#define	NUM_S_TILES	16
#define	NUM_T_TILES	16
#else
#define	TILE_S_SIZE	127
#define	TILE_T_SIZE	127
#define	NUM_S_TILES	11
#define	NUM_T_TILES	11
#endif
#define	TEX_S_SIZE	(num_s_tiles*TILE_S_SIZE)
#define	TEX_T_SIZE	(num_t_tiles*TILE_T_SIZE)

#define	MAX_S_COORD	(TEX_S_SIZE - TILE_S_SIZE)
#define	MAX_T_COORD	(TEX_T_SIZE - TILE_T_SIZE)
#define	SQR2		1.4142135623730950488
#define	UTIL_S_SIZE\
	((rotationAngle!=0.) ? (float)MAX_S_COORD/SQR2 :  (float)MAX_S_COORD)
#define	UTIL_T_SIZE\
	((rotationAngle!=0.) ? (float)MAX_T_COORD/SQR2 :  (float)MAX_T_COORD)

static GLuint *textureId, *xtile, *ytile;
static GLboolean *isObsolete, *isResident;

#define	tileNdx(a,b)\
	(((b)%num_t_tiles)*num_s_tiles+((a)%num_s_tiles))

static u_char *picture;
static char *msg[] =
{
    "Scale & Bias",
    "3x3 Edge Convolution",
    "3x3 Sharpen Convolution",
    "7x7 Blur Convolution"
};

int MakeFilterKernel( GLfloat, GLfloat, int );

static float
	rotationAngle = 0., cosAngle=1., sinAngle=0.,
	zoomFactor = 1.,
	origf[2];

static GLfloat xfactor=1., yfactor=0.;

static int
	num_s_tiles, num_t_tiles,
	orig[2], maxorig[2],
	autopilot = 0, autodir = 0,
	filterMode = 0, textured = GL_TRUE,
	shrinkSize = 0,
	xsize, ysize, zsize,
	W, H, TX0=0, TY0=0,
	last_minx, last_maxx, last_miny, last_maxy;

static struct {
    int     lastx, lasty,
            middledown, rightdown, leftdown;
    float   dx, dy;
} mouse;

static void Reshape( int w, int h );
static void Redraw( void );
static void Mouse( int, int, int, int );
static void Motion( int, int );
static void Key( unsigned char, int, int );
static void Special( int, int, int );

static void drawNoTex( void );
static void drawSubTex( void );
static void drawTiles( void );

/*---------------------------------------------------------------------------*/
/*	InitRoamDisplay
/*---------------------------------------------------------------------------*/
int
InitRoamDisplay( char *filename, int x, int y, int w, int h )
{
	int i, j, win;
	short *rbuf, *gbuf, *bbuf;
	IMAGE *image;
	u_char *dataptr;

	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE );
	glutInitWindowPosition( x, y );
	glutInitWindowSize( w, h );
	win = glutCreateWindow( "roam" );
	glutDisplayFunc( Redraw );
	glutSpecialFunc( Special );
	glutReshapeFunc( Reshape );
	glutMouseFunc( Mouse );
	glutMotionFunc( Motion );
	glutKeyboardFunc( Key );

	if ((image = iopen( filename, "rw" )) == NULL)
   	{
		fprintf( stderr, "ERROR::file %s not found\n", filename );
		exit(-1);
   	}

	rbuf = (short *) malloc( image->xsize*sizeof(short));
	gbuf = (short *) malloc( image->xsize*sizeof(short));
	bbuf = (short *) malloc( image->xsize*sizeof(short));

	xsize = (((image->xsize-1)/TILE_S_SIZE) + 1)*TILE_S_SIZE;
	ysize = (((image->ysize-1)/TILE_T_SIZE) + 1)*TILE_T_SIZE;
	zsize = (image->zsize==1) ? 1 : 4;

	num_s_tiles = INF((xsize/TILE_S_SIZE), NUM_S_TILES);
	num_t_tiles = INF((ysize/TILE_T_SIZE), NUM_T_TILES);

	if (num_s_tiles != num_t_tiles)
	{
		xsize = ysize = SUP(xsize,ysize);
		num_s_tiles = num_t_tiles = SUP(num_s_tiles,num_t_tiles);
	}

/*
	Actually, we need to allocate 7 more pixels, because :
	1) we may transfer (TILE_S_SIZE+1, TILE_T_SIZE+1) sized tiles into
	texture objects.
	2) we may use a 7-wide convolution filter kernel (blur), shrinking 6 pixels
*/
	xsize += 7;
	ysize += 7;

	picture = (u_char *) calloc( 1, xsize*ysize*zsize );
	for (i=0; i<image->ysize; i++)
	{
		if (zsize==1)
		{
			getrow( image, rbuf, i, 0 );
			for (j=0; j<image->xsize; j++)
			{
				picture[i*xsize+j] = (u_char) rbuf[j];
			}
		}
		else
		{
			getrow( image, rbuf, i, 0 );
			getrow( image, gbuf, i, 1 );
			getrow( image, bbuf, i, 2 );
			for (j=0; j<image->xsize; j++)
			{
				dataptr = picture + 4*(i*xsize + j);
				dataptr[0] = 255;
				dataptr[1] = (u_char) bbuf[j];
				dataptr[2] = (u_char) gbuf[j];
				dataptr[3] = (u_char) rbuf[j];
			}
		}
	}
	iclose( image );

	textureId = (GLuint *) malloc( num_s_tiles*num_t_tiles*sizeof(GLuint) );
	xtile = (GLuint *) malloc( num_s_tiles*num_t_tiles*sizeof(GLuint) );
	ytile = (GLuint *) malloc( num_s_tiles*num_t_tiles*sizeof(GLuint) );
	isResident = (GLboolean *)malloc(num_s_tiles*num_t_tiles*sizeof(GLboolean));
	isObsolete = (GLboolean *)malloc(num_s_tiles*num_t_tiles*sizeof(GLboolean));

	return win;
}

/*---------------------------------------------------------------------------*/
/*	InitGraphique
/*---------------------------------------------------------------------------*/
static void
InitGraphique( void )
{
	int i, j;

	glEnable( GL_TEXTURE_2D );
#ifdef	SUBTEX
	glTexImage2D(
		GL_TEXTURE_2D,
		0,
		(zsize==1) ? GL_LUMINANCE8_EXT : GL_RGB8_EXT,
		TEX_S_SIZE, TEX_T_SIZE,
		0,
		(zsize==1) ? GL_LUMINANCE : GL_ABGR_EXT,
		GL_UNSIGNED_BYTE,
		NULL
	);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
#else	/* SUBTEX */
	glGenTexturesEXT( num_s_tiles*num_t_tiles, textureId );
	for (i=0; i<num_s_tiles; i++)
	for (j=0; j<num_t_tiles; j++)
	{
		glBindTextureEXT( GL_TEXTURE_2D, textureId[tileNdx(i,j)] );
		glTexImage2D(
			GL_TEXTURE_2D,
			0,
			(zsize==1) ? GL_LUMINANCE8_EXT : GL_RGB8_EXT,
			TILE_S_SIZE+1, TILE_T_SIZE+1,
			0,
			(zsize==1) ? GL_LUMINANCE : GL_ABGR_EXT,
			GL_UNSIGNED_BYTE,
			NULL
		);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	}
#endif	/* SUBTEX */

	maxorig[0] = ((xsize-TEX_S_SIZE)/TILE_S_SIZE)*TILE_S_SIZE;
	maxorig[1] = ((ysize-TEX_T_SIZE)/TILE_T_SIZE)*TILE_T_SIZE;

	rotationAngle = 0.;
	zoomFactor = 1.;
	origf[0] = origf[1] = 0.;
	orig[0] = orig[1] = 0;
	last_minx = -1;
	last_miny = -1;
	last_maxx = -1;
	last_maxy = -1;

	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	glClearColor( 0., 0., 0., 1. );
	glClear( GL_COLOR_BUFFER_BIT );
	glutSwapBuffers();
}

/*---------------------------------------------------------------------------*/
/*      Special
/*---------------------------------------------------------------------------*/
static void
Special( int key, int x, int y )
{
	switch (key)
	{
		case GLUT_KEY_LEFT:
			autopilot = 1;
			autodir = LEFTWARD;
			break;

		case GLUT_KEY_RIGHT:
			autopilot = 1;
			autodir = RIGHTWARD;
			break;

		case GLUT_KEY_UP:
			autopilot = 1;
			autodir = UPWARD;
			break;

		case GLUT_KEY_DOWN :
			autopilot = 1;
			autodir = DOWNWARD;
			break;
	}
	glutIdleFunc( autopilot ? Redraw : NULL );
	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*      Reshape
/*---------------------------------------------------------------------------*/
static void
Reshape( int w, int h )
{
	static int first = 1;
	glViewport( 0, 0, w, h );
	W = w;
	H = h;
	if (first)
	{
		InitGraphique();
		first = 0;
	}
}

/*---------------------------------------------------------------------------*/
/*	Mouse
/*---------------------------------------------------------------------------*/
static void Mouse( int button, int state, int x, int y )
{
	mouse.lastx = x;
	mouse.lasty = y;
	mouse.dx = 0.;
	mouse.dy = 0.;
	switch (button)
	{
		case GLUT_LEFT_BUTTON:
			mouse.leftdown = (state==GLUT_DOWN);
			break;

		case GLUT_RIGHT_BUTTON:
			mouse.rightdown = (state==GLUT_DOWN);
			break;

		case GLUT_MIDDLE_BUTTON:
			mouse.middledown = (state==GLUT_DOWN);
			break;
	}
	if (mouse.leftdown || mouse.rightdown || mouse.middledown)
	{
		autopilot = 0;
		glutIdleFunc( NULL );
	}
	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Motion
/*---------------------------------------------------------------------------*/
static void
Motion( int x, int y )
{
	mouse.dx = x-mouse.lastx;
	mouse.dy = y-mouse.lasty;
	mouse.lastx = x;
	mouse.lasty = y;

	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Key
/*---------------------------------------------------------------------------*/
static void
Key( unsigned char key, int x, int y )
{
	switch (key)
	{
		case 27:	/* Escape */
			exit(0);

		case 'k':
		case 'K':
			filterMode = (filterMode+1)%4;
			last_minx = -1;	/* FORCE ALL TILES UPDATE */
			last_maxx = -1;
			last_miny = -1;
			last_maxy = -1;
			break;

		case 'r':
		case 'R':
			rotationAngle = 0.;
			zoomFactor = 1.;
			break;
	}
	textured = GL_TRUE;
	glutPostRedisplay();
}


/*---------------------------------------------------------------------------*/
/*	MARK_TILES
/*---------------------------------------------------------------------------*/
static void
MARK_TILES( int xf, int xl, int yf, int yl, int xd, int yd )
{
	int i, j,
		numx = xl-xf+1,
		numy = yl-yf+1;

	for (i=0; i<numx; i++)
	for (j=0; j<numy; j++)
	{
		int ndx = tileNdx(xd+i,yd+j);
		isObsolete[ndx] = 1;
		xtile[ndx] = xf+i;
		ytile[ndx] = yf+j;
	}
}

/*---------------------------------------------------------------------------*/
/*	checkTiles
/*---------------------------------------------------------------------------*/
static void
checkTiles( int xorg, int yorg )
{
	int	minx, miny,
		maxx, maxy,
		dx, dy;

	minx = (xorg<0) ? 0 : (xorg/TILE_S_SIZE);
	miny = (yorg<0) ? 0 : (yorg/TILE_T_SIZE);
	minx = INF( minx, (maxorig[0]/TILE_S_SIZE) );
	miny = INF( miny, (maxorig[1]/TILE_T_SIZE) );

	maxx = minx + num_s_tiles - 1;
	maxy = miny + num_t_tiles - 1;
	if ((minx > last_maxx) || (maxx < last_minx) ||
		(miny > last_maxy) || (maxy < last_miny))
	{
		MARK_TILES( minx, maxx, miny, maxy, TX0, TY0 );
		last_minx = minx;
		last_maxx = maxx;
		last_miny = miny;
		last_maxy = maxy;
		return;
	}

	dx = minx-last_minx;
	dy = miny-last_miny;

	if (dx > 0)
	{
		MARK_TILES( maxx, maxx, last_miny, last_maxy-TY0, TX0, TY0 );
		MARK_TILES( maxx, maxx, last_maxy-TY0+1, last_maxy, TX0, 0 );
		TX0 = (TX0+1)%num_s_tiles;
	}
	else if (dx < 0)
	{
		TX0--;
		if (TX0 < 0) TX0+=num_s_tiles;
		MARK_TILES( minx, minx, last_miny, last_maxy-TY0, TX0, TY0 );
		MARK_TILES( minx, minx, last_maxy-TY0+1, last_maxy, TX0, 0 );
	}
	if (dy > 0)
	{
		MARK_TILES( minx, maxx-TX0, maxy, maxy, TX0, TY0 );
		MARK_TILES( maxx-TX0+1, maxx, maxy, maxy, 0, TY0 );
		TY0 = (TY0+1)%num_t_tiles;
	}
	else if (dy < 0)
	{
		TY0--;
		if (TY0 < 0) TY0+=num_s_tiles;
		MARK_TILES( minx, maxx-TX0, miny, miny, TX0, TY0 );
		MARK_TILES( maxx-TX0+1, maxx, miny, miny, 0, TY0 );
	}
	
	last_minx = minx;
	last_miny = miny;
	last_maxx = maxx;
	last_maxy = maxy;
}

/*---------------------------------------------------------------------------*/
/*	TRANSLATE_ORIGIN
/*---------------------------------------------------------------------------*/
static void
TRANSLATE_ORIGIN( int x, int y )
{
	float dx, dy;

	dx = (float)(x)*cosAngle - (float)(y)*sinAngle;
	dy = (float)(y)*cosAngle + (float)(x)*sinAngle;
	dx *= ((float)TEX_S_SIZE/1024.)/zoomFactor;
	dy *= ((float)TEX_T_SIZE/1024.)/zoomFactor;
	dx = INF( SUP(dx, -TILE_S_SIZE), TILE_S_SIZE );
	dy = INF( SUP(dy, -TILE_T_SIZE), TILE_T_SIZE );

	origf[0] -= dx;
	origf[1] += dy;
	orig[0] = (int)origf[0];
	orig[1] = (int)origf[1];
}

/*---------------------------------------------------------------------------*/
/*	Redraw
/*---------------------------------------------------------------------------*/
static void
Redraw( void )
{
	int i, do_clear=0;
	float mins, maxs, mint, maxt;

/*
	we limit the refresh rate to 30HZ, to avoid jerking between 30-60HZ frames
*/
	glXSwapIntervalSGI( autopilot ? 2 : 1 );

	if (autopilot)
	{
		float du=0., dv=0.;
		switch (autodir)
		{
			case RIGHTWARD: du = -ARROW_STRIDE; break;
			case LEFTWARD : du =  ARROW_STRIDE; break;
			case UPWARD   : dv =  ARROW_STRIDE; break;
			case DOWNWARD : dv = -ARROW_STRIDE; break;
		}
		TRANSLATE_ORIGIN( du, dv );
		textured = GL_TRUE;
	}
	else if (mouse.middledown)
	{
		if (mouse.leftdown)
		{
			int dy=-mouse.dy;
			while (dy>0) { zoomFactor -= zoomFactor*.005; dy--; }
			while (dy<0) { zoomFactor += zoomFactor*.005; dy++; }
			if (zoomFactor < 1.) zoomFactor = 1.;
			textured = GL_TRUE;
		}
/*
		We allow user-control of the filter parameters only when
		the pixel size drawn with glDrawPixels() is equal to
		the pixel size drawn with textured polygons
		This depends mostly on the window size & aspect ratio
*/
#ifndef SUBTEX
		else if ((mouse.dx || mouse.dy) &&
			 (W== 899) && (H==899))
		{
			xfactor += mouse.dx/150.0;
			yfactor -= mouse.dy/150.0;
			textured = GL_FALSE;
		}
#endif	/*SUBTEX */
	}
	else if (mouse.rightdown)
	{
		float angle;
		rotationAngle -= .5*(float)mouse.dy;
		angle = DEG2RAD(rotationAngle);
		sinAngle = fsin(angle);
		cosAngle = fcos(angle);
		textured = GL_TRUE;
	}
	else if (mouse.leftdown)
	{
		TRANSLATE_ORIGIN( mouse.dx, mouse.dy );
		textured = GL_TRUE;
	}

#ifndef SUBTEX
/*
	Blur filter has no tuning parameters in this demo
*/
	if ((filterMode < 3) &&
		(W==899) && (H==899) &&
		(!textured))
	{
		drawNoTex();

		rotationAngle = 0.;
		cosAngle = 1.;
		sinAngle = 0.;
		zoomFactor = 1.;

		last_minx = -1;	/* FORCE ALL TILES UPDATE */
		last_maxx = -1;
		last_miny = -1;
		last_maxy = -1;

		return;
	}
#endif

	checkTiles( orig[0], orig[1] );

/*
	Projection matrix settings
	--------------------------
*/
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	mins = (float)MAX_S_COORD*(.5-.25*SQR2);
	maxs = (float)MAX_S_COORD*(.5+.25*SQR2);
	mint = (float)MAX_T_COORD*(.5-.25*SQR2);
	maxt = (float)MAX_T_COORD*(.5+.25*SQR2);
	glOrtho( mins, maxs, mint, maxt, -1., 1. );
	glScalef( 1., (float)W/(float)H, 1.);

/*
	A rough optimization is done to avoid unnecessary clear
	Could be finer tuned in case of (zoomFactor<1.)
*/
	if (  do_clear  || (zoomFactor < 1.) ||
		(orig[0]<0) || (orig[0]>maxorig[0]+TILE_S_SIZE) ||
		(orig[1]<0) || (orig[1]>maxorig[1]+TILE_T_SIZE) )
		glClear( GL_COLOR_BUFFER_BIT );

#ifdef	SUBTEX
	drawSubTex();
#else	/* SUBTEX */
	drawTiles();
#endif	/* SUBTEX */

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0., (float)W, 0., (float)H, -1., 1. );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

    glRasterPos2i( 10, 10 );
    glColor3f( 1., 1., 1. );
    for (i=0; i<strlen(msg[filterMode]); i++)
        glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24, msg[filterMode][i] );
    glRasterPos2i( 0, 0 );

	glutSwapBuffers();
}

/*---------------------------------------------------------------------------*/
/*	drawNoTex
/*---------------------------------------------------------------------------*/
static void
drawNoTex( void )
{
	int i, w, h;
	float x, y;

	glDisable( GL_TEXTURE_2D );
	shrinkSize = MakeFilterKernel(xfactor, yfactor, filterMode);

/*
	indicate unpacking parameters
*/
	x = orig[0] + (float)MAX_S_COORD*(.5-.25*SQR2);
	y = orig[1] + (float)MAX_T_COORD*(.5-.25*SQR2);

	glPixelStorei( GL_UNPACK_ROW_LENGTH, xsize );
	glPixelStorei( GL_UNPACK_SKIP_PIXELS, x );
	glPixelStorei( GL_UNPACK_SKIP_ROWS, y );

	w = (x+W < xsize ) ? W : xsize-x;
	h = (y+H < ysize ) ? H : ysize-y;

	if (x<0) w+=x;
	if (y<0) h+=y;

	if ((w!=W) || (h!=H))
		glClear( GL_COLOR_BUFFER_BIT );

	glRasterPos2i( -INF(x,0), -INF(y,0) );
	glDrawPixels(
		w+shrinkSize, h+shrinkSize,
		(zsize==1) ? GL_LUMINANCE : GL_ABGR_EXT,
		GL_UNSIGNED_BYTE,
		picture
	);
	glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
	glPixelStorei( GL_UNPACK_SKIP_PIXELS, 0 );
	glPixelStorei( GL_UNPACK_SKIP_ROWS, 0 );
	glDisable( GL_CONVOLUTION_2D_EXT );
	glDisable( GL_SEPARABLE_2D_EXT );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0., (float)W, 0., (float)H, -1., 1. );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

    glRasterPos2i( 10, 10 );
    glColor3f( 1., 1., 1. );
    for (i=0; i<strlen(msg[filterMode]); i++)
        glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24, msg[filterMode][i] );
    glRasterPos2i( 0, 0 );

	glutSwapBuffers();
}

/*---------------------------------------------------------------------------*/
/*	drawSubTex
/*---------------------------------------------------------------------------*/
static void
drawSubTex( void )
{
	int i, j;

/*
	ModelView matrix settings
	-------------------------
*/
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	glTranslatef( .5*(float)MAX_S_COORD, .5*(float)MAX_T_COORD, 0. );
	glRotatef( rotationAngle, 0., 0., 1. );
	glScalef( zoomFactor, zoomFactor, 1. );
	glTranslatef( -.5*(float)MAX_S_COORD, -.5*(float)MAX_T_COORD, 0. );

	if (orig[0]<0)
		glTranslatef((float)(-orig[0]),0.,0.);
	else if (orig[0]>=maxorig[0]+TILE_S_SIZE)
		glTranslatef((float)(maxorig[0]+TILE_S_SIZE-orig[0]),0.,0.);

	if (orig[1]<0)
		glTranslatef(0.,(float)(-orig[1]),0.);
	else if (orig[1]>=maxorig[1]+TILE_T_SIZE)
		glTranslatef(0.,(float)(maxorig[1]+TILE_T_SIZE-orig[1]),0.);

/*
	Texture Matrix settings
	-----------------------
*/
	glMatrixMode( GL_TEXTURE );
	glLoadIdentity();

	glScalef( 1./(float)TEX_S_SIZE, 1./(float)TEX_T_SIZE, 1. );
	if (orig[0]>maxorig[0]+TILE_S_SIZE)
		glTranslatef((float)(maxorig[0]+TILE_S_SIZE), 0., 0. );
	else if (orig[0]>0)
		glTranslatef((float) orig[0], 0., 0. );
	if (orig[1]>maxorig[1]+TILE_T_SIZE)
		glTranslatef( 0., (float)(maxorig[1]+TILE_T_SIZE), 0. );
	else if (orig[1]>0)
		glTranslatef( 0., (float)orig[1], 0. );
	glMatrixMode( GL_MODELVIEW );

/*
	update new tiles
	----------------
*/
	shrinkSize = MakeFilterKernel( xfactor, yfactor, filterMode );

	glEnable( GL_TEXTURE_2D );
	glPixelStorei( GL_UNPACK_ROW_LENGTH, xsize );

	for (i=0; i<num_s_tiles; i++)
	for (j=0; j<num_t_tiles; j++)
	{
		int ndx = tileNdx(i+TX0,j+TY0);
		if (isObsolete[ndx])
		{
			glPixelStorei( GL_UNPACK_SKIP_PIXELS, xtile[ndx]*TILE_S_SIZE );
			glPixelStorei( GL_UNPACK_SKIP_ROWS, ytile[ndx]*TILE_T_SIZE );

			glTexSubImage2DEXT(
				GL_TEXTURE_2D,
				0,
				((i+TX0)%num_s_tiles)*TILE_S_SIZE,
				((j+TY0)%num_t_tiles)*TILE_T_SIZE,
				TILE_S_SIZE+shrinkSize,
				TILE_T_SIZE+shrinkSize,
				(zsize==1) ? GL_LUMINANCE : GL_ABGR_EXT,
				GL_UNSIGNED_BYTE,
				picture
			);
			isObsolete[ndx] = GL_FALSE;
		}
	}
	glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
	glPixelStorei( GL_UNPACK_SKIP_PIXELS, 0 );
	glPixelStorei( GL_UNPACK_SKIP_ROWS, 0 );
	glDisable( GL_CONVOLUTION_2D_EXT );
	glDisable( GL_SEPARABLE_2D_EXT );
/*
	ENFIN: Redraw
	-------------
*/
#define TV2I(x,y) glTexCoord2i((x),(y)); glVertex2i((x),(y));

	glBegin( GL_TRIANGLE_STRIP );
		TV2I( 0, 0 );
		TV2I( 0, MAX_T_COORD );
		TV2I( MAX_S_COORD, 0 );
		TV2I( MAX_S_COORD, MAX_T_COORD );
	glEnd();

}

/*---------------------------------------------------------------------------*/
/*	drawTiles
/*---------------------------------------------------------------------------*/
static void
drawTiles( void )
{
	int i, j;

/*
	ModelView matrix settings
	-------------------------
*/
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	glTranslatef( .5*(float)MAX_S_COORD, .5*(float)MAX_T_COORD, 0. );
	glRotatef( rotationAngle, 0., 0., 1. );
	glScalef( zoomFactor, zoomFactor, 1. );
	glTranslatef( -.5*(float)MAX_S_COORD, -.5*(float)MAX_T_COORD, 0. );

	if (orig[0]<0)
		glTranslatef((float)(-orig[0]),0.,0.);
	else if (orig[0]>=maxorig[0]+TILE_S_SIZE)
		glTranslatef((float)(maxorig[0]-orig[0]),0.,0.);
	else
		glTranslatef(-(float)(orig[0]%TILE_S_SIZE),0.,0.);

	if (orig[1]<0)
		glTranslatef(0.,(float)(-orig[1]),0.);
	else if (orig[1]>=maxorig[1]+TILE_T_SIZE)
		glTranslatef(0.,(float)(maxorig[1]-orig[1]),0.);
	else
		glTranslatef(0.,-(float)(orig[1]%TILE_T_SIZE),0.);

/*
	Texture Matrix settings
	-----------------------
*/
	glMatrixMode( GL_TEXTURE );
	glLoadIdentity();
	glScalef( 1./(float)(TILE_S_SIZE+1), 1./(float)(TILE_T_SIZE+1), 1. );
	glMatrixMode( GL_MODELVIEW );

/*
	ENFIN: redraw
	-------------
*/
	glEnable( GL_TEXTURE_2D );
	glAreTexturesResidentEXT(
		num_s_tiles*num_t_tiles,
		textureId,
		isResident
	);

/*
	FIRST: draw all tiles resident in TRAM
	--------------------------------------
*/
	for (i=0; i<num_s_tiles; i++)
	for (j=0; j<num_t_tiles; j++)
	{
		int ndx = tileNdx(i+TX0,j+TY0);
		if ((!isObsolete[ndx]) && isResident[ndx])
		{
			float	x0 = i*TILE_S_SIZE + .5,
					y0 = j*TILE_T_SIZE + .5,
					x1 = x0 + TILE_S_SIZE,
					y1 = y0 + TILE_T_SIZE;

			glBindTextureEXT( GL_TEXTURE_2D, textureId[ndx] );
			glBegin( GL_TRIANGLE_STRIP );
			glTexCoord2f( .5, .5 ); glVertex2f(x0,y0);
			glTexCoord2f( .5, TILE_T_SIZE+.5); glVertex2f(x0,y1);
			glTexCoord2f( TILE_S_SIZE+.5, .5); glVertex2f(x1,y0);
			glTexCoord2f( TILE_S_SIZE+.5, TILE_T_SIZE+.5); glVertex2f(x1,y1);
			glEnd();
		}
	}
/*
	THEN: restore all tiles non-resident in TRAM, if needed
	-------------------------------------------------------
*/
	for (i=0; i<num_s_tiles; i++)
	for (j=0; j<num_t_tiles; j++)
	{
		int ndx = tileNdx(i+TX0,j+TY0);
		if ((!isObsolete[ndx]) && !isResident[ndx])
		{
			float	x0 = i*TILE_S_SIZE + .5,
					y0 = j*TILE_T_SIZE + .5,
					x1 = x0 + TILE_S_SIZE,
					y1 = y0 + TILE_T_SIZE;

			glBindTextureEXT( GL_TEXTURE_2D, textureId[ndx] );
			glBegin( GL_TRIANGLE_STRIP );
			glTexCoord2f( .5, .5 ); glVertex2f(x0,y0);
			glTexCoord2f( .5, TILE_T_SIZE+.5); glVertex2f(x0,y1);
			glTexCoord2f( TILE_S_SIZE+.5, .5); glVertex2f(x1,y0);
			glTexCoord2f( TILE_S_SIZE+.5, TILE_T_SIZE+.5); glVertex2f(x1,y1);
			glEnd();
		}
	}
/*
	FINALLY: load & draw newest tiles
	---------------------------------
*/
	shrinkSize = MakeFilterKernel( xfactor, yfactor, filterMode );
	shrinkSize++;	/* we manage our tiling ourselves */

	glPixelStorei( GL_UNPACK_ROW_LENGTH, xsize );
	for (i=0; i<num_s_tiles; i++)
	for (j=0; j<num_t_tiles; j++)
	{
		int ndx = tileNdx(i+TX0,j+TY0);
		if (isObsolete[ndx])
		{
			glPixelStorei( GL_UNPACK_SKIP_PIXELS, xtile[ndx]*TILE_S_SIZE );
			glPixelStorei( GL_UNPACK_SKIP_ROWS, ytile[ndx]*TILE_T_SIZE );

			glBindTextureEXT( GL_TEXTURE_2D, textureId[ndx] );
			glTexSubImage2DEXT(
				GL_TEXTURE_2D,
				0,
				0, 0,
				TILE_S_SIZE+shrinkSize, TILE_T_SIZE+shrinkSize,
				(zsize==1) ? GL_LUMINANCE : GL_ABGR_EXT,
				GL_UNSIGNED_BYTE,
				picture
			);
		}
	}
	glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
	glPixelStorei( GL_UNPACK_SKIP_PIXELS, 0 );
	glPixelStorei( GL_UNPACK_SKIP_ROWS, 0 );
	glDisable( GL_CONVOLUTION_2D_EXT );
	glDisable( GL_SEPARABLE_2D_EXT );
	for (i=0; i<num_s_tiles; i++)
	for (j=0; j<num_t_tiles; j++)
	{
		int ndx = tileNdx(i+TX0,j+TY0);
		if (isObsolete[ndx])
		{
			float	x0, y0, x1, y1;

			x0 = i*TILE_S_SIZE + .5;
			y0 = j*TILE_T_SIZE + .5;
			x1 = x0 + TILE_S_SIZE;
			y1 = y0 + TILE_T_SIZE;

			glBindTextureEXT( GL_TEXTURE_2D, textureId[ndx] );
			glBegin( GL_TRIANGLE_STRIP );
			glTexCoord2f( .5, .5 ); glVertex2f(x0,y0);
			glTexCoord2f( .5, TILE_T_SIZE+.5); glVertex2f(x0,y1);
			glTexCoord2f( TILE_S_SIZE+.5, .5); glVertex2f(x1,y0);
			glTexCoord2f( TILE_S_SIZE+.5, TILE_T_SIZE+.5); glVertex2f(x1,y1);
			glEnd();

			isObsolete[ndx] = 0;
		}
	}
}
